/*******************************************************************\

Module: A CNF Clause Container

Author: CM Wintersteiger

\*******************************************************************/

#include <errno.h>
#include <assert.h>

#include <algorithm>
#include <iostream>
#include <sstream>

#include "cnf.h"
#include "exception.h"

Literal::Literal( unsigned v, bool sign )
{
  set(v, sign);
}

Literal::Literal( signed s )
{
  set((s<0)?-s:s, (s<0));
}

signed long long Literal::to_dimacs( void ) const
{
  if (sign())
    return - ((signed long long) variable());
  else
    return variable();
}

Literal DClause::contains( unsigned v ) const
{
  for (unsigned i=0; i<size(); i++)
    if (content[i]==v)
      return content[i];
  return Literal(0,false);
}

bool DClause::contains( const Literal &l ) const
{
  for (unsigned i=0; i<size(); i++)
    if (content[i]==l)
      return true;
  return false;
}

void DClause::remove( unsigned v )
{
  unsigned i=0;
  while (i<size())
  {
    if (content[i]==v)
    {
      content[i] = content.back();
      content.pop_back();
      #ifdef __THERE_IS_ENOUGH_MEMORY
        count--;
      #endif
    }
    else
      i++;
  }
}

void DClause::remove( const Literal &l )
{
  unsigned i=0;
  while (i<size())
  {
    if (content[i]==l)
    {
      content[i] = content.back();
      content.pop_back();
      #ifdef __THERE_IS_ENOUGH_MEMORY
        count--;
      #endif
    }
    else
      i++;
  }
}

bool DClause::insert( const Literal &l )
{
  if (!contains(l))
  {
    content.resize(size()+1, l);
    #ifdef __THERE_IS_ENOUGH_MEMORY
      count++;
    #endif
    return true;
  }
  else
    return false;
}

bool DClause::is_true( void ) const
{
  for (unsigned i=0; i<size(); i++)
  {
    for (unsigned j=i+1; i<size(); j++)
      if (content[i].variable() == content[j].variable() &&
          content[i].sign() != content[j].sign())
        return true;
  }
  return false;
}

bool DClause::empty( void ) const
{
  return content.empty();
}

void DClause::add( const DClause& c )
{
  content.reserve(size()+c.size());
  for (unsigned i=0; i<c.size(); i++)
    insert(c.content[i]);
}

void DClause::add_except( const DClause& c, unsigned exception )
{
  content.reserve(size()+c.size());
  for (unsigned i=0; i<c.size(); i++)
    if (c.content[i].variable()!=exception)
      insert(c.content[i]);
}

void DClause::clear( void )
{
  content.clear();
  #ifdef __THERE_IS_ENOUGH_MEMORY
    count = 0;
  #endif
}

void DClause::reduce(
  const std::vector<unsigned> &levels,
  const std::vector<Quantifier> &quantifiers)
{  
  Quantifier::Type max_type = Quantifier::UNIVERSAL;
  unsigned max_ql = 1;
  unsigned max_var = 0;

  while (max_ql>0 && max_type == Quantifier::UNIVERSAL)
  {
    max_ql = 0;
    max_var = 0;

    for (unsigned i=0; i<size(); i++)
    {
      unsigned lvl = levels[content[i].variable()];
      if (lvl > max_ql)
      {
        max_ql = lvl;
        max_type = quantifiers[content[i].variable()].type;
        max_var = content[i].variable();
      }
    }
    
    if (max_var!=0 && max_type==Quantifier::UNIVERSAL)
    {
      //std::cout << "Reducing " << max_var << std::endl;
      remove(max_var);
    }
  }
}

bool DClause::qresolve_into(
  const DClause &other,
  unsigned pivot,
  const std::vector<unsigned> &levels,
  const std::vector<Quantifier> &quantifiers,
  DClause &into ) const
{
  Literal lc1 = contains(pivot);
  Literal lc2 = other.contains(pivot);

  into.clear();

  if (  lc1.variable()==lc2.variable() &&
        lc1.sign() != lc2.sign() )
  {
    into.add_except(*this, pivot);
    into.add_except(other, pivot);
    into.reduce(levels, quantifiers);
    //std::cout << " [" << size() << "]" << std::endl;
    return true;
  }
  return false;
}

std::string DClause::to_string( void ) const
{
  std::stringstream result;
  result << "(";
  unsigned i = 0;
  while (i<size())
  {
    result << content[i].to_dimacs();
    i++;
    if (i!=size())
      result << ", ";
  }
  result << ")";
  return result.str();
}

bool DClause::invert( CClause& cube ) const
{
  cube.clear();
  for (unsigned i=0; i<size(); i++)
  {
    Literal l = content[i];
    l.invert();
    cube.insert(l);
  }
  return true;
}

bool QCNF::qresolve( unsigned ci1, unsigned ci2, unsigned v)
{
  if ( ci1 >= size() || ci2 >= size() ) return false;

  DClause c1 = at(ci1);
  DClause c2 = at(ci2);

  std::cout << "Resolving [" << ci1+1 << "," << ci2+1 << "] " << c1.to_string() <<
    " x " << c2.to_string() << " over " << v << std::endl;

  push_back(DClause());
  if ( c1.qresolve_into(  c2, v,
                          quantification_levels,
                          quantifiers, back()))
    return true;

  pop_back();
  //std::cout << " [N/A]" << std::endl;
  return false;
}

std::string CClause::to_string( void ) const
{
  std::stringstream result;
  result << "[";
  const_iterator it=content.begin();
  while (it!=content.end())
  {
    result << it->to_dimacs();
    it++;
    if (it!=content.end())
      result << ", ";
  }
  result << "]";
  return result.str();
}

bool CClause::is_false( void ) const
{
  for (const_iterator it=content.begin();
       it!=content.end();
       it++)
  {
    for (const_iterator it2=it;
         it2!=content.end();
         it2++)
    {
      if (Literal(it->variable(), ! it->sign()) == *it2)
        return true;
    }
  }
  return false;
}

void CClause::add( const CClause& c )
{
  for (const_iterator it=c.content.begin();
       it!=c.content.end();
       it++)
  {
    insert(*it);
  }
}

bool CClause::invert( DClause& clause ) const
{
  clause.clear();
  for (const_iterator it = content.begin();
       it!=content.end();
       it++)
  {
    Literal l = *it;
    l.invert();
    clause.insert(l);
  }
  return true;
}

void CNF::output( std::ostream& stream ) const
{
  stream << "[" << size() << "]";
  for (std::vector<DClause>::const_iterator it = begin();
       it!=end();
       it++)
  {
    stream << it->to_string();
  }
}

void DNF::output( std::ostream& stream ) const
{
  stream << "[" << size() << "]";
  for (std::vector<CClause>::const_iterator it = begin();
       it!=end();
       it++)
  {
    stream << it->to_string();
  }
}

bool CNF::invert( DNF& dnf ) const
{
  //std::cout << "DEBUG: Inverting CNF:";
  //output(std::cout); std::cout << std::endl;
  dnf.clear();

  for (const_iterator it = begin();
       it!=end();
       it++)
  {
    CClause t;
    if (!it->invert(t)) return false;
    dnf.push_back(t);
  }
  //std::cout << "DEBUG: Result DNF:";
  //dnf.output(std::cout); std::cout << std::endl;
  return true;
}

bool DNF::invert( CNF &cnf ) const
{
//  std::cout << "DEBUG: Inverting DNF:";
//  output(std::cout); std::cout << std::endl;
  cnf.clear();

  for (const_iterator it = begin();
       it!=end();
       it++)
  {
    DClause t;
    if (!it->invert(t)) return false;
    cnf.push_back(t);
  }
//  std::cout << "DEBUG: Result CNF:";
//  cnf.output(std::cout); std::cout << std::endl;
  return true;
}

/*******************************************************************\

Function: QCNF::read_qdimacs

  Inputs: stream

 Outputs: nothing

 Purpose: loads a qdimacs file into a cnf, throws exceptions on error.

\*******************************************************************/

bool QCNF::read_qdimacs( std::ifstream &in )
{
  if (!in)
  {
    last_error = "QBF File not open.";
    return false;
  }

  std::string t;
  in >> t;
  while (t=="c")
  {
    char c=0x0;
    while (c != 0xA) in.read(&c,1);
    in >> t;
  }

  if (t!="p")
  {
    last_error = "QBF File does not start with a p line.";
    return false;
  }
  in >> t;
  if (t!="cnf")
  {
    last_error = "QBF File does not have a cnf argument.";
    return false;
  }
  in >> t;
  errno = 0;
  max_variable_no = strtoul(t.c_str(), NULL, 10);
  if (errno!=0)
  {
    last_error = "QBF File: Number of variables is bogus.";
    return false;
  }
  in >> t;
  errno = 0;
  unsigned noclauses = strtoul(t.c_str(), NULL, 10);
  if (errno!=0)
  {
    last_error = "QBF File: Number of clauses is bogus.";
    return false;
  }
  reserve(noclauses);
  in >> t;
  Quantifier::Type cur_type = Quantifier::NONE;

  while (t=="e" || t=="a")
  {
    if (t=="e") cur_type = Quantifier::EXISTENTIAL;
    else cur_type = Quantifier::UNIVERSAL;
    errno = 0;
    unsigned varno=1;
    while (varno!=0)
    {
      in >> t;
      varno = strtoul(t.c_str(), NULL, 10);
      if (errno!=0)
      {
        last_error = "QBF File: Invalid quantification sequence.";
        return false;
      }
      if (varno!=0)
      {
        new_variable(varno, cur_type);
      }
    }
    in >> t;
  }

  DClause tcl;
  signed v;
  char *endp;

  errno = 0;
  while (in && size()<noclauses)
  {
    v = strtol(t.c_str(), &endp, 10);
    if (errno!=0 || !(*t.c_str()!='\0' && *endp=='\0'))
    {
      last_error = "QBF File: Bogus clause.";
      return false;
    }

    if (v==0)
    {
      push_back(tcl);      
      tcl.clear();
    }
    else if (v>0)
    {
      tcl.insert(Literal(v, false));
    }
    else // v<0
    {
      tcl.insert(Literal(-v, true));
    }

    // quantify unquantified variables
    if (v!=0)
    {
      if (v<0) v=-v;
      if ((unsigned)v>max_variable_no) max_variable_no = v;
      if ((unsigned)v>=quantifiers.size()) quantifiers.resize(v+1);
      if (quantifiers[v].type==Quantifier::NONE)
      {
        new_variable(v, Quantifier::EXISTENTIAL, true);
      }
    }

    in >> t;
  }

  return true;
}

void QCNF::write_qdimacs( std::ofstream& out ) const
{
  out << "p cnf " << max_variable_no << " " << size() << std::endl;
  for (std::vector<qscopet>::const_iterator it=quantification_sequence.begin();
       it!=quantification_sequence.end();
       it++)
  {
    for (qscopet::const_iterator sit = it->begin();
         sit!=it->end();
         sit++)
    {
      if (sit->type==Quantifier::EXISTENTIAL)
        out << "e " << sit->variable << " 0" << std::endl;
      else if (sit->type==Quantifier::UNIVERSAL)
        out << "a " << sit->variable << " 0" << std::endl;
    }
  }
  for (CNF::const_iterator it = begin();
       it!=end();
       it++)
  {
    for (DClause::const_iterator lit = it->begin();
         lit!=it->end();
         lit++)
    {
      out << lit->to_dimacs() << " ";
    }
    out << "0" << std::endl;
  }
}

bool QCNF::new_variable( unsigned varno, Quantifier::Type t, bool outer )
{
  if (outer && t==Quantifier::UNIVERSAL)
  {
    last_error = "New outermost variables may not be universal.";
    return false;
  }

  if (is_defined(varno))
  {
    last_error = "Variable redefined";
    return false;
  }

  if (varno>=quantifiers.size())
    quantifiers.resize(varno+1);
  quantifiers[varno].type = t;
  quantifiers[varno].variable = varno;

  if (outer || quantification_sequence.size()==0)
  {
    if (varno>=quantification_levels.size())
      quantification_levels.resize(varno+1, 0);
    if (t==Quantifier::UNIVERSAL)
    {
      quantification_levels[varno] = 1; // leave some space for outermost ex.
      if (quantification_sequence.size()<2)
        quantification_sequence.resize(2);
      quantification_sequence[1].push_front(quantifiers[varno]);
    }
    else
    {
      quantification_levels[varno] = 0;
      if (quantification_sequence.size()==0)
        quantification_sequence.resize(1);
      quantification_sequence[0].push_front(quantifiers[varno]);
    }

  }
  else
  {
    qscopet& last = quantification_sequence.back();

    if (varno>=quantification_levels.size())
      quantification_levels.resize(varno+1, 0);
    if (last.back().type==t)
    {
      quantification_levels[varno] =
        quantification_levels[last.back().variable];
      last.push_back(quantifiers[varno]);
    }
    else
    {
      quantification_levels[varno] =
        quantification_levels[last.back().variable]+1;
      quantification_sequence.resize(quantification_sequence.size()+1);
      quantification_sequence.back().push_back(quantifiers[varno]);
    }

  }

  if (varno>max_variable_no) max_variable_no = varno;

//  std::cout << "DEBUG: new innermost variable #" << varno << ", ql=" <<
//    quantification_levels[varno] << std::endl;

  return true;
}

bool QCNF::new_variable_after(
  unsigned varno,
  Quantifier::Type t,
  unsigned after)
{
  if (!is_defined(after))
  {
    std::stringstream s;
    s << "Definition depends on undefined variable #" << after;
    last_error = s.str();
    return false;
  }
  if (is_defined(varno))
  {
    last_error = "Variable redefinition";
    return false;
  }

  unsigned newlvl = quantification_levels[after];
  qscopet &scope = quantification_sequence[newlvl];

  Quantifier q(t, varno);

  if (scope.front().type==t)
  {
    scope.push_back(q); // we don't care about the order in the scope
  }
  else
  {
    newlvl++;
    if (quantification_sequence.size()<=newlvl)
      quantification_sequence.resize(quantification_sequence.size()+1);
    quantification_sequence[newlvl].push_back(q);
  }

  if (varno>=quantification_levels.size())
    quantification_levels.resize(varno+1, 0);
  quantification_levels[varno] = newlvl;

  if (varno>=quantifiers.size())
    quantifiers.resize(varno+1);
  quantifiers[varno] = q;

  if (varno>max_variable_no) max_variable_no = varno;

//  std::cout << "DEBUG: new variable #" << varno << ", ql=" <<
//    quantification_levels[varno] << std::endl;

  return true;
}

bool QCNF::is_defined( unsigned v ) const
{
   if (v < quantifiers.size() &&
       quantifiers[v].type!=Quantifier::NONE)
       return true;
   return false;
}

bool QCNF::is_universal( unsigned v ) const
{
  if (v < quantifiers.size() &&
      quantifiers[v].type==Quantifier::UNIVERSAL)
      return true;
 return false;
}

bool QCNF::is_before( unsigned v1, unsigned v2 ) const
{
  if (quantification_levels[v1] < quantification_levels[v2])
    return true;
  else if (quantification_levels[v1] > quantification_levels[v2])  
    return false;
  
  assert(quantification_levels[v1] == quantification_levels[v2]);
  
  return false;

#if false
  // precise check, but this is not necessary, since we have a cyclic dependency
  // check afterwards.  
  for(std::vector<qscopet>::const_iterator it=quantification_sequence.begin();
      it!=quantification_sequence.end();
      it++)
  {
    qscopet::const_iterator sit=it->begin();
    for( ; sit!=it->end(); sit++)
      if(sit->variable==v1 || sit->variable==v2) break;
    
    if(sit!=it->end()) // found v1 or v2 in the *it scope
    {
      if(sit->variable==v1)
        return true;
      else
      {
        assert(sit->variable==v2);
        return false;
      }
      
      break;
    }
  }
  
  // we should never get here.
  assert(false);
  return false;
#endif
}

bool QCNF::check_fresh(const Literal &l) const
{
  if (is_defined(l.variable()))
  {
    std::stringstream ss;
    ss << "Variable is not fresh: " << l.to_dimacs();
    throw Exception(ss.str());
  }
  
  return true;
}

bool QCNF::check_defined(const Literal &l) const
{
  if (!is_defined(l.variable()))
  {
    std::stringstream ss;
    ss << "Undefined variable: " << l.to_dimacs();
    throw Exception(ss.str());
  }
  
  return true;
}

void QCNF::show_quantification_sequence(std::ostream &out) const
{
  out << "Quantification Sequence: "; 
  for (std::vector<QCNF::qscopet>::const_iterator it =
         quantification_sequence.begin();
       it!=quantification_sequence.end();
       it++)
  {
    if(it->size()>0)
    {
      if(is_universal(it->front().variable))
        out << " A ";
      else
        out << " E ";
    }
    
    for(QCNF::qscopet::const_iterator qit=it->begin();
        qit!=it->end();
        qit++)
    {
      if(qit!=it->begin()) out << ",";
      out << qit->variable;
    }
  }
  out << std::endl;
}

bool QCNF::apply_extension( const ite_extension &ext )
{     
  if(ext.new_var.sign())
  {
    last_error = "Negated extensions are not supported.";
    return false;
  }
  
  check_fresh(ext.new_var);
  check_defined(ext.condition);
  check_defined(ext.true_branch);
  check_defined(ext.false_branch);

  // find quantification level of new variable
  unsigned maxlvl = quantification_levels[ext.condition.variable()];
  Quantifier maxq =  quantifiers[ext.condition.variable()];
  if (quantification_levels[ext.true_branch.variable()] > maxlvl)
  {
    maxlvl = quantification_levels[ext.true_branch.variable()];
    maxq = quantifiers[ext.true_branch.variable()];
  }
  if (quantification_levels[ext.false_branch.variable()] > maxlvl)
  {
    maxlvl = quantification_levels[ext.false_branch.variable()];
    maxq = quantifiers[ext.false_branch.variable()];
  }
  if (!new_variable_after(
          ext.new_var.variable(), Quantifier::EXISTENTIAL, maxq.variable) )
    return false;      

  CNF f;
  DClause c1, c2, c3, c4; // 4 clauses will be added to the qcnf.
  c1.insert(Literal(ext.new_var.variable(), true));
  c1.insert(ext.true_branch);
  c2.insert(Literal(ext.new_var.variable(), true));
  c2.insert(ext.condition);
  c2.insert(ext.false_branch);
  c3.insert(Literal(ext.new_var.variable(), false));
  c4.insert(Literal(ext.new_var.variable(), false));
  c4.insert(ext.condition);
  
  Literal invcond = ext.condition; invcond.invert(); 
  Literal invtrue = ext.true_branch; invtrue.invert(); 
  Literal invfalse= ext.false_branch; invfalse.invert();
  c1.insert(invcond);
  c3.insert(invcond);
  c3.insert(invtrue);
  c4.insert(invfalse);
  push_back(c1);
  push_back(c2);
  push_back(c3);
  push_back(c4);
  
  return true;
}

bool QCNF::apply_extension( const and_extension &ext )
{
  if(ext.new_var.sign())
  {
    last_error="Negated extensions are not supported.";
    return false;
  }
    
  check_fresh(ext.new_var);
  for(std::vector<Literal>::const_iterator it=ext.literals.begin();
      it!=ext.literals.end();
      it++)
    check_defined(*it);
  
    
  DClause bigone;
  bigone.insert(ext.new_var);
    
  Quantifier maxq;
  unsigned maxlvl;
      
  if(ext.literals.size()>0)
  {
    maxq=quantifiers[ext.literals.front().variable()];
    maxlvl = quantification_levels[ext.literals.front().variable()];
    Literal invnew = ext.new_var; invnew.invert();
    
    for(std::vector<Literal>::const_iterator it=ext.literals.begin();
        it!=ext.literals.end();
        it++)
    {
      Literal l=*it;

      DClause c;
      c.insert(invnew);
      c.insert(l);
      push_back(c);      
      
      bigone.insert(~l);

      if (quantification_levels[l.variable()] > maxlvl)
      {
        maxlvl = quantification_levels[l.variable()];
        maxq = quantifiers[l.variable()];
      }
      
      // std::cout << c.to_string() << std::endl;
    }
  }
  
  push_back(bigone); // ... + 1 clauses will be added to the qcnf.

  // std::cout << bigone.to_string() << std::endl;

  if ( ext.literals.size()==0 ) // doesn't depend on anything
  {
    if (! new_variable( ext.new_var.variable(), Quantifier::EXISTENTIAL, true ) )
      return false;
  }
  else
  {
    if (! new_variable_after(
          ext.new_var.variable(), Quantifier::EXISTENTIAL, maxq.variable) )
      return false;
  }
  
  return true;
}

bool QCNF::is_applicable( const and_extension &ext )
{  
  for(std::vector<Literal>::const_iterator lit=ext.literals.begin();
      lit!=ext.literals.end();
      lit++)
    if(!is_defined(lit->variable()))
      return false;
  
  return true;
}

bool QCNF::is_applicable( const ite_extension &ext )
{
  return is_defined(ext.condition.variable()) &&
         is_defined(ext.true_branch.variable()) &&
         is_defined(ext.false_branch.variable());
}
