/*******************************************************************\

Module: A CNF Clause Container

Author: CM Wintersteiger

\*******************************************************************/

#ifndef CNF_H_
#define CNF_H_

#include <vector>
#include <list>
#include <iostream>
#include <fstream>

#include "literal.h"
#include "extensions.h"

#define __THERE_IS_ENOUGH_MEMORY

//forward decls
class CClause;
class DNF;
class CNF;

// The Quantifier Class
class Quantifier {
  private:
  public:
    enum Type { NONE, UNIVERSAL, EXISTENTIAL };
    Type type;
    unsigned variable;
    Quantifier( Type t, unsigned v ) { type = t; variable = v; };
    Quantifier( void ) : type(NONE), variable(0) {};
};

// The Clause Class
class DClause {
  protected:
    std::vector<Literal> content;
    #ifdef __THERE_IS_ENOUGH_MEMORY
      unsigned count;
    #endif
    
  public:
    DClause( void ) 
      #ifdef __THERE_IS_ENOUGH_MEMORY
        : count(0) 
      #endif
      {};
    
    Literal contains( unsigned ) const;
    bool contains( const Literal& ) const;
    
    void remove( unsigned );
    void remove( const Literal& );
    
    bool insert( const Literal& );
    
    bool is_true( void ) const;
    bool is_false( void ) const { return false; }
    bool empty( void ) const;
    #ifdef __THERE_IS_ENOUGH_MEMORY
      inline unsigned size( void ) const { return count; };
    #else
      inline unsigned size( void ) const { return content.size(); };
    #endif
    void clear( void );
    
    void add( const DClause& );
    void add_except( const DClause&, unsigned );
    bool invert( CClause& ) const;
    
    void reduce( const std::vector<unsigned> &, 
                 const std::vector<Quantifier> &);
    bool qresolve_into( const DClause&, 
                        unsigned,
                        const std::vector<unsigned> &,
                        const std::vector<Quantifier> &, 
                        DClause &) const;                                      
                 
    std::string to_string( void ) const;
    
    typedef std::vector<Literal>::const_iterator const_iterator;
    typedef std::vector<Literal>::iterator iterator;
    
    const_iterator begin( void ) const
      { return content.begin(); }
    const_iterator end( void ) const
      { return content.end(); }
};

// The Cube Class
class CClause : public DClause {
  public:
    void add( const CClause& );
    bool invert( DClause& ) const;
    bool is_true( void ) const { return false; }
    bool is_false( void ) const;
    std::string to_string( void ) const;
    
    bool expand( CNF& ) const;
};

// The CNF Class
class CNF : public std::vector<DClause> {
  public:
    unsigned max_variable_no;
    
    bool invert( DNF& ) const;
    void output( std::ostream& ) const;    
};

// The DNF Class
class DNF : public std::vector<CClause> {
  public:
    unsigned max_variable_no;
  
    bool invert( CNF& ) const;
    void output( std::ostream& ) const;
};

// The QCNF Class
class QCNF : public CNF {
  private:
  public:
    std::string last_error;
        
    std::vector<unsigned> quantification_levels;        
    std::vector<Quantifier> quantifiers;
    typedef std::list<Quantifier> qscopet;
    std::vector<qscopet> quantification_sequence;
    
    bool new_variable( unsigned, Quantifier::Type, bool outer=false );
    bool new_variable_after( unsigned, Quantifier::Type, unsigned );
    
    bool is_defined( unsigned ) const;
    bool is_universal( unsigned v ) const;
    bool is_before( unsigned v1, unsigned v2 ) const;
    
    bool check_fresh(const Literal &) const;
    bool check_defined(const Literal &) const;
    
    bool qresolve( unsigned, unsigned, unsigned );
    
    bool apply_extension( const and_extension & );
    bool apply_extension( const ite_extension & );
    bool is_applicable( const and_extension & );
    bool is_applicable( const ite_extension & );
    
    bool read_qdimacs( std::ifstream & );
    void write_qdimacs( std::ofstream & ) const;
    
    void show_quantification_sequence(std::ostream &out) const;
};

#endif

