/*******************************************************************\

Module: QBV, The Quantified Boolean Validator

Author: CM Wintersteiger

\*******************************************************************/

#include <errno.h>

#include <set>
#include <list>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>

#ifdef __USE_MINISAT2
  #include <SimpSolver.h>
#else
  #include <Solver.h>
#endif

#include <sys/time.h>
#include <time.h>

#include "qbv.h"
#include "exception.h"

/*******************************************************************\

Function: QBV::QBV

  Inputs: none

 Outputs: none

 Purpose: Constructor.

\*******************************************************************/

QBV::QBV( void )
{
  cur_linenumber=1;
  cur_colnumber=1;
  test_trace = false;
  debug = false;
}

/*******************************************************************\

Function: QBV::validate

  Inputs: char *, char *

 Outputs: true if the certificate was correct, false otherwise

 Purpose: validates a certificate file.

\*******************************************************************/

bool QBV::validate( const char *qbf, const char* qbc )
{
  std::string qbffn = qbf;
  std::string qbcfn = qbc;
  return validate(qbffn, qbcfn);
}

/*******************************************************************\

Function: QBV::read_resolution_line

  Inputs: certificate file object

 Outputs: nothing

 Purpose: processes a resolution line

\*******************************************************************/

void QBV::read_resolution_line( std::ifstream &certificate_file )
{
  unsigned c1 = read_unsigned(certificate_file);
  unsigned c2 = read_unsigned(certificate_file);
  unsigned v = read_unsigned(certificate_file);

  if (!qcnf.qresolve(c1-1, c2-1, v))
    throw Exception("Resolution failed.");
}

/*******************************************************************\

Function: QBV::read_trace_line

  Inputs: certificate file object

 Outputs: nothing

 Purpose: processes a resolution line

\*******************************************************************/

void QBV::read_trace_line( unsigned newinx, std::ifstream &certificate_file )
{
  Resolution resolution;

  resolution.newinx = newinx;

  if (newinx<clause_map.size() && clause_map[newinx]!=0)
  {
    std::stringstream s;
    s << "Redefinition of clause #" << newinx;
    throw Exception(s.str());
  }

  // parse resolvent
  while (certificate_file.peek()==' ') read_character(certificate_file);
  if (certificate_file.peek()!='*') // we have a resolvent in there.
  {
    Literal l = read_literal(certificate_file);
    while (l.variable()!=0)
    {
      resolution.resolvent.insert(l);
      if (!qcnf.is_defined(l.variable()))
        qcnf.new_variable(l.variable(), Quantifier::EXISTENTIAL, true);
      l = read_literal(certificate_file);
    }
  }
  else
  {
    std::string& t = read_token(certificate_file);
    if (t=="*")
      resolution.star_resolvent = true;
    else
      throw Exception("Bogus resolvent.");
  }

  // parse antecedents
  bool all_antecedents_here=true;
  unsigned a = read_unsigned(certificate_file);
  while (a!=0)
  {
    resolution.antecedents.push_back(a);

    if (a>=clause_map.size() || clause_map[a]==0)
      all_antecedents_here=false;

    a = read_unsigned(certificate_file);
  }

  if(all_antecedents_here)
    apply_resolution(resolution);
  else
    resolvents_list.push_front(resolution);

}

/*******************************************************************\

Function: QBV::apply_resolution

  Inputs: resolution object, bool

 Outputs: nothing

 Purpose: actually performs the resolution provided

\*******************************************************************/

void QBV::apply_resolution( Resolution &resolution )
{
  // check resolution chain
  if (resolution.antecedents.empty()) // this is a root clause
  {
    if (!ignore_no_reason)
      throw Exception("Resolvent not justified (no antecedents).");
    qcnf.push_back(resolution.resolvent);
    if (resolution.newinx>=clause_map.size())
      clause_map.resize(resolution.newinx+1, 0);
    clause_map[resolution.newinx] = qcnf.size(); // this is 1 too high!
  }
  else if (resolution.antecedents.size()==1)
  {
    throw Exception("Single antecedent does not make sense.");
  }
  else if (!resolution.star_resolvent)
  {
//    if (debug)
//      std::cout << "DEBUG: " << resolution.newinx << ": supposed resolvent " <<
//        resolution.resolvent.to_string() << std::endl;

    static std::list<unsigned> pivots;
    static DClause tclause, rclause;

    pivots.clear();
    order_antecedents(resolution.resolvent, resolution.antecedents, pivots);

    std::list<unsigned>::const_iterator it = resolution.antecedents.begin();
    tclause = qcnf[clause_map[*it] - 1];
    it++;

    bool toggler = false;
    for (std::list<unsigned>::const_iterator pit = pivots.begin();
         it!=resolution.antecedents.end() && pit!=pivots.end();
         it++, pit++)
    {
      if (!
        qcnf[clause_map[*it]-1].qresolve_into(
          (toggler)?rclause:tclause,
          *pit,
          qcnf.quantification_levels,
          qcnf.quantifiers,
          (toggler)?tclause:rclause))
        throw Exception("Resolution failed.");

      toggler = !toggler;
    }
    DClause &final = (toggler)?rclause:tclause;

//    if (debug)
//      std::cout << "DEBUG: " << resolution.newinx << ": " <<
//        final.to_string() << std::endl;

    qcnf.push_back(final);
    if (resolution.newinx>=clause_map.size())
      clause_map.resize(resolution.newinx+1, 0);
    clause_map[resolution.newinx] = qcnf.size(); // this is 1 too high!

//    if (debug)
//      std::cout << "DEBUG: Inserted " << resolution.newinx << " as " <<
//        qcnf.size()-1 << std::endl;

    // check that the given resolvent is equal
    if (resolution.resolvent.size()!=final.size())
      throw Exception("Calculated and supplied resolvent sizes do not match.");
    for (DClause::const_iterator it=final.begin();
         it!=final.end();
         it++)
    {
      if (!resolution.resolvent.contains(*it))
        throw Exception("Calculated and supplied resolvents differ.");
    }
  }
  else // *-resolvent
  {
    qcnf.push_back(DClause());
    DClause &resolvent = qcnf.back();
    std::list<unsigned> doubles;

    for (std::list<unsigned>::const_iterator it =
           resolution.antecedents.begin();
         it!=resolution.antecedents.end();
         it++)
    {
      DClause &a = qcnf[clause_map[*it]-1];

      for (DClause::const_iterator cit = a.begin();
           cit!=a.end();
           cit++)
      {
        Literal l = *cit; l.invert();

        if (resolvent.contains(l))
        {
          doubles.push_back(l.variable());
        }

        resolvent.insert(*cit);
      }
    }

    if (doubles.size()<resolution.antecedents.size()-1)
      throw Exception("Not enough pivot variables for resolution.");
    else if (doubles.size()>resolution.antecedents.size()-1)
    {
      std::stringstream s;
      s << "Tautological resolvent encountered: " << resolvent.to_string();
      throw Exception(s.str());
    }

    for (std::list<unsigned>::const_iterator it = doubles.begin();
         it!=doubles.end();
         it++)
    {
      resolvent.remove(*it);
    }

    resolvent.reduce(qcnf.quantification_levels, qcnf.quantifiers);

//    if (debug)
//      std::cout << resolution.newinx << ": " << resolvent.to_string() <<
//        std::endl;

    if (resolution.newinx>=clause_map.size())
      clause_map.resize(resolution.newinx+1, 0);
    clause_map[resolution.newinx] = qcnf.size(); // this is 1 too high!
  }
}

/*******************************************************************\

Function: QBV::read_extension_line

  Inputs: new variable number, certificate file object

 Outputs: nothing

 Purpose: processes an extension line

\*******************************************************************/

void QBV::read_extension_line( std::ifstream &certificate_file )
{
  Literal newlit = read_literal(certificate_file);

  qcnf.check_fresh(newlit);

  std::string &t = read_token(certificate_file);
  if (t=="I")
  {
    ite_extension ext=read_extension_ite(newlit.variable(), certificate_file);
    if(!apply_extension_ite(ext))
      ite_ext_list.push_back(ext);
  }
  else if (t=="A")
  {
    and_extension ext=read_extension_and(newlit.variable(), certificate_file);
    if(!apply_extension_and(ext))
      and_ext_list.push_back(ext);
  }
  else
    throw Exception("Unknown extension type supplied.");
}

/*******************************************************************\

Function: QBV::read_extension_ite

  Inputs: certificate file object

 Outputs: nothing

 Purpose: processes an ITE-extension line

\*******************************************************************/

ite_extension QBV::read_extension_ite(
  unsigned newvar,
  std::ifstream &certificate_file )
{
  ite_extension ext;

  ext.line_no = cur_linenumber;
  ext.new_var = Literal(newvar, false);
  ext.condition = read_literal(certificate_file);
  ext.true_branch = read_literal(certificate_file);
  ext.false_branch = read_literal(certificate_file);

  return ext;
}

/*******************************************************************\

Function: QBV::extension_ite

  Inputs: certificate file object

 Outputs: nothing

 Purpose: applies an ITE-extension

\*******************************************************************/

bool QBV::apply_extension_ite( ite_extension &ext )
{
  if(qcnf.is_applicable(ext))
  {
    qcnf.apply_extension(ext);
    
    model_cache_ite[ext.new_var.variable()].push_back(ext.condition);
    model_cache_ite[ext.new_var.variable()].push_back(ext.true_branch);
    model_cache_ite[ext.new_var.variable()].push_back(ext.false_branch);

    return true;
  }

  return false;
}

/*******************************************************************\

Function: QBV::read_extension_and

  Inputs: new variable number, certificate file object

 Outputs: nothing

 Purpose: processes an AND-extension line

\*******************************************************************/

and_extension QBV::read_extension_and(
  unsigned newvar,
  std::ifstream &certificate_file )
{
  and_extension ext;

  ext.line_no = cur_linenumber;
  ext.new_var = Literal(newvar, false);

  signed x=read_signed(certificate_file);
  while(x!=0)
  {
    Literal l((x<0)?-x:x, (x<0));
    ext.literals.push_back(l);
    x = read_signed(certificate_file);
  }

  return ext;
}

/*******************************************************************\

Function: QBV::extension_and

  Inputs: new variable number, certificate file object

 Outputs: nothing

 Purpose: applies an AND-extension

\*******************************************************************/

bool QBV::apply_extension_and( and_extension &ext )
{
  if(qcnf.is_applicable(ext))
  {
    qcnf.apply_extension(ext);
    
    if(ext.literals.size()==0)
      model_cache_and[ext.new_var.variable()].clear();
    else
    {
      std::vector<Literal> &cache=model_cache_and[ext.new_var.variable()];
      cache.reserve(ext.literals.size());

      for(std::vector<Literal>::const_iterator it=ext.literals.begin();
          it!=ext.literals.end();
          it++)
      {
        const Literal &l=*it;
        cache.push_back(l);
      }
    }

    return true;
  }

  return false;
}

/*******************************************************************\

Function: QBV::read_conclusion_line

  Inputs: certificate file object

 Outputs: nothing

 Purpose: processes a conclusion line

\*******************************************************************/

bool QBV::read_conclusion_line( std::ifstream &certificate_file )
{
  if(debug)
    std::cout << "Concluding ..." << std::endl;

  apply_delayed_resolutions();
  check_delayed_resolutions();

  apply_delayed_extensions();
  check_delayed_extensions();

  std::string &t = read_token(certificate_file);
  if (t=="INVALID" || t=="invalid")
  {
    unsigned ci = read_unsigned(certificate_file);
    DClause final_clause = qcnf.at(clause_map[ci]-1);
    final_clause.reduce(qcnf.quantification_levels, qcnf.quantifiers);
    //std::cout << "Final clause: " << final_clause.to_string() << std::endl;
    return (final_clause.size()==0);
  }
  else if (t=="VALID" || t=="valid");
  {
    model_type model;

    if (last_original_clause==(unsigned)-1)
      throw Exception("A model cannot be validated without a QBF file.");

    while (!certificate_file.eof())
    {
      unsigned exvar;
      
      try { exvar=read_unsigned(certificate_file); }
      catch (...) { break; }
      
      if(exvar > last_original_variable)
      {
        std::stringstream s;
        s << "Variable " << exvar << " does not need a model.";
        throw Exception(s.str());
      }
      
      signed mlit = read_signed(certificate_file);
      read_whitespace(certificate_file);

      if(!qcnf.is_defined(Literal(mlit).variable()))
      {
        std::stringstream s;
        s << "Model for " << exvar << " (" << mlit << ") is undefined."; 
        throw Exception(s.str());
      }
        
      model[exvar] = mlit;
    }

    check_model_quantification(model);
    check_function_dependencies(model);

    return test_clausewise( model );
  }
}

/*******************************************************************\

Function: QBV::validate

  Inputs: string, string

 Outputs: true if the certificate was correct, false otherwise

 Purpose: validates a certificate file.

\*******************************************************************/

bool QBV::validate( const std::string &qbf, const std::string &qbc )
{
  gettimeofday(&starttime, NULL);

  if (test_trace) return validate_trace(qbc);

  std::string t;
  std::ifstream certificate_file(qbc.c_str());

  if (!certificate_file.is_open())
    throw Exception("Cannot open file: '" + qbc +"'.");

  t = read_token(certificate_file);
  if (t!="QBCertificate")
    throw Exception("Missing header in file: '" + qbc + "'.");


  if (!qbf.empty())
  {
    std::ifstream qbf_file(qbf.c_str());
    if (!qbf_file)
      throw Exception("Couldn't open file " + qbf);
    if (!qcnf.read_qdimacs(qbf_file))
      throw Exception(qcnf.last_error);
  }

  clause_map.resize(qcnf.size()+1, 0);
  for (unsigned i=1; i<=qcnf.size(); i++)
  {
    clause_map[i]=i;
  }

  last_original_clause = qcnf.size() - 1;
  last_original_variable = qcnf.max_variable_no;

  if (debug)
  {
    unsigned qcount = 0;
    for (std::vector<QCNF::qscopet>::const_iterator it =
           qcnf.quantification_sequence.begin();
         it!=qcnf.quantification_sequence.end();
         it++)
      qcount += it->size();
    std::cout << qcount << " quantifiers and " <<
      qcnf.size() << " clauses loaded." << std::endl;

    std::cout << "Applying resolution and extension steps ..." << std::endl;
  }

  while (certificate_file)
  {
    t = read_token(certificate_file);
    if (t=="C" || t=="c") // Comment line
    {
      while ( !certificate_file.eof() &&
              read_character(certificate_file)!='\n');
    }
    else if (t=="R" || t=="r") // Resolution line
      read_resolution_line(certificate_file);
    else if (t=="E" || t=="e") // Extension Rules
      read_extension_line(certificate_file);
    else if (t=="CONCLUDE" || t=="conclude")
      return read_conclusion_line(certificate_file);
    else
    {
      // assume trace-style
      errno=0;
      char *endp;
      unsigned val = strtoul(t.c_str(), &endp, 10);
      if (errno!=0 || !(*t.c_str()!='\0' && *endp=='\0'))
      {
        std::stringstream s;
        s << "Number conversion failed (" << t << ").";
        throw Exception(s.str());
      }
      read_trace_line(val, certificate_file);
    }
  }

  // At this point we should have read a conclusion line already
  throw Exception("Certificate incomplete.");
}

/*******************************************************************\

Function: QBV::test_clausewise

  Inputs: model

 Outputs: true if the certificate was correct, false otherwise

 Purpose: checks if the provided model actually is a model.

\*******************************************************************/

bool QBV::test_clausewise( model_type &model )
{
  #ifdef __USE_MINISAT2
    SimpSolver solver;
  #else
    Solver solver;
  #endif

  struct timeval before;
  gettimeofday(&before, NULL);

  while ((unsigned) solver.nVars()<=qcnf.max_variable_no)
    solver.newVar();

  // add function definitions
  for (unsigned i=last_original_clause+1; i<qcnf.size(); i++)
  {
    vec<Lit> c;
    for (DClause::const_iterator lit = qcnf[i].begin();
         lit!=qcnf[i].end();
         lit++)
    {
      c.push( Lit(lit->variable(), lit->sign()) );
    }
    solver.addClause(c);
  }

  // add model
  for (model_type::const_iterator mit = model.begin();
       mit!=model.end();
       mit++)
  {
    vec<Lit> c1, c2;
    c1.push( Lit(mit->first, false) );
    c1.push( Lit((mit->second<0)?-mit->second:mit->second, !(mit->second<0)) );

    c2.push( Lit(mit->first, true) );
    c2.push( Lit((mit->second<0)?-mit->second:mit->second, (mit->second<0)) );

    solver.addClause(c1);
    solver.addClause(c2);
  }

  #ifdef __USE_MINISAT2
    solver.eliminate();
  #endif

  // test every clause.
  for (unsigned i=0; i<=last_original_clause; i++)
  {
    vec<Lit> assumptions;

    if (debug)
      std::cout << "\rTesting clause " << i+1 << "/" << last_original_clause+1;
    
//    unsigned highest_order=0;
    for (DClause::const_iterator lit = qcnf[i].begin();
         lit!=qcnf[i].end();
         lit++)
    {
      assumptions.push( Lit( lit->variable(), ! lit->sign() ) );
      
//      unsigned v=lit->variable();
//      unsigned i=0;
//      for(; i<function_order.size(); i++)
//        if(function_order[i]==v)
//          break;
//      if(i>highest_order) highest_order=i;
    }
    
//    for(unsigned j=highest_order+1; j<function_order.size(); j++)
//    {
//      if(qcnf.is_universal(function_order[j]))
//        assumptions.push( Lit(function_order[j], false) );
//      else if(function_order[j]<=last_original_variable &&
//              model.find(function_order[j])==model.end())
//        assumptions.push( Lit(function_order[j], false) );
//    }

    if (!solver.okay()) throw Exception("Minisat not ready.");

    if (solver.solve(assumptions))
    {
      std::cout << std::endl << "Counterexample found." << std::endl;
      std::cout << "Clause (#" << i+1 << "): " << qcnf[i].to_string() << std::endl;
      if (debug)
      {
        std::cout << "Assignment to universals: ";
        for (unsigned i=1; i<qcnf.max_variable_no; i++)
        {
          if(qcnf.is_universal(i))
            std::cout << i << "=" << (solver.model[i]==l_True) << " ";
        }
        std::cout << std::endl;
      }
      return false;
    } // else all fine.
  }

  if (debug)
  {
    struct timeval after;
    gettimeofday(&after, NULL);

    std::cout << std::endl;

    long mseconds;
    mseconds=(after.tv_sec-before.tv_sec)*1000;
    mseconds+=(after.tv_usec-before.tv_usec)/1000;
    std::cout << "SAT TIME: " << mseconds/1000.0 << " sec." << std::endl;

    long tmseconds;
    tmseconds=(after.tv_sec-starttime.tv_sec)*1000;
    tmseconds+=(after.tv_usec-starttime.tv_usec)/1000;
    std::cout << "TOT TIME: " << tmseconds/1000.0 << " sec." << std::endl;
    std::cout << "SAT PERCENTAGE: " << 100*(mseconds/(double)tmseconds) << " %" << std::endl;
  }

  return true; // no counterexample found
}

/*******************************************************************\

Function: QBV::read_unsigned

  Inputs: stream

 Outputs: unsigned

 Purpose: reads a token from a stream and tries to convert it to an
          unsigned. Throws exceptions on errors.

\*******************************************************************/

unsigned QBV::read_unsigned( std::ifstream &in )
{
  std::string &t = read_token(in);
  //std::cout << "DEBUG: Converting " << t << std::endl;
  errno=0;
  char *endp;
  unsigned val = strtoul(t.c_str(), &endp, 10);
  if (errno!=0 || !(*t.c_str()!='\0' && *endp=='\0'))
  {
    std::stringstream s;
    s << "Number conversion failed (" << t << ").";
    throw Exception(s.str());
  }
  return val;
}

/*******************************************************************\

Function: QBV::read_signed

  Inputs: stream

 Outputs: signed

 Purpose: reads a token from a stream and tries to convert it to a
          signed integer. Throws exceptions on errors.

\*******************************************************************/

signed QBV::read_signed( std::ifstream &in )
{
  std::string &t = read_token(in);
  //std::cout << "DEBUG: Converting " << t << std::endl;
  errno=0;
  char *endp;
  signed val = strtol(t.c_str(), &endp, 10);
  if (errno!=0 || !(*t.c_str()!='\0' && *endp=='\0'))
  {
    std::stringstream s;
    s << "Number conversion failed (" << t << ").";
    throw Exception(s.str());
  }
  return val;
}

/*******************************************************************\

Function: QBV::read_literal

  Inputs: stream

 Outputs: signed

 Purpose: reads a signed integer from a stream and converts it to a
          literal. Throws exceptions on errors.

\*******************************************************************/

Literal QBV::read_literal( std::ifstream &in )
{
  signed int x = read_signed(in);
  return Literal( (x<0)?-x:x, (x<0) );
}

/*******************************************************************\

Function: QBV::read_token

  Inputs: stream

 Outputs: string

 Purpose: reads a token from a stream. Throws exceptions on errors.

\*******************************************************************/

std::string& QBV::read_token( std::ifstream &in )
{
  char c;

  cur_token.clear();
  // pre-white
  c = in.peek();
  while ((c==' ' || c=='\n' || c=='\r') && !in.eof())
  {
    read_character(in);
    c = in.peek();
  }
  // token
  while (c!=' ' && c!='\n' && c!='\r' && !in.eof())
  {
    cur_token+=read_character(in);
    c = in.peek();
  }

  //std::cout << "DEBUG: Read '" << cur_token << "'." << std::endl;

  return cur_token;
}

/*******************************************************************\

Function: QBV::read_character

  Inputs: stream

 Outputs: char

 Purpose: reads a character from a stream. Throws exceptions on errors.

\*******************************************************************/

char QBV::read_character( std::ifstream &in )
{
  char c;
  in.get(c);
  if (c=='\n' || c=='\r')
  {
    cur_linenumber++;
    cur_colnumber=1;
  }
  else
    cur_colnumber++;
  return c;
}

/*******************************************************************\

Function: QBV::read_whitespace

  Inputs: stream

 Outputs: nothing

 Purpose: eats whitespace characters

\*******************************************************************/

void QBV::read_whitespace( std::ifstream &in )
{
  char c = in.peek();
  while (!in.eof() && (c==' ' || c=='\n' || c=='\r' || c=='\t'))
  {
    read_character(in);
    c=in.peek();
  }
}

/*******************************************************************\

Function: QBV::validate_trace

  Inputs: string

 Outputs: true if the trace is correct, false otherwise.

 Purpose: validates a resolution trace.

\*******************************************************************/

bool QBV::validate_trace( const std::string &filename )
{
  std::ifstream trace_file(filename.c_str());

  clause_map.clear();

  while (!trace_file.eof())
  {
    read_whitespace(trace_file);
    if (trace_file.eof()) break;

    read_trace_line(read_unsigned(trace_file), trace_file);
  }

  if (qcnf.back().size()==0) // the last one must be empty then.
    return true;

  trace_file.close();
  return false;
}

/*******************************************************************\

Function: QBV::order_antecedents

  Inputs: resolvent and antecedents

 Outputs: nothing

 Purpose: orders the antecedents so that trivial resolution can be done.

\*******************************************************************/

void QBV::order_antecedents(
  const DClause &resolvent,
  std::list<unsigned> &antecedents,
  std::list<unsigned> &pivots ) const
{
//  if (debug)
//    std::cout << "DEBUG: ordering: " << antecedents.size() <<
//      " antecedents." << std::endl;

  DClause assumptions = resolvent; // copy
  std::vector<DClause> clauses;
  std::vector<unsigned> clause_indexes;

  unsigned checksize = antecedents.size();

  //std::cout << "DEBUG: old order: ";
  for (std::list<unsigned>::const_iterator it = antecedents.begin();
       it!=antecedents.end();
       it++)
  {
    if (*it>=clause_map.size() || clause_map[*it]==0)
      throw Exception("Inconsistent clause map.");
    clauses.push_back(qcnf.at(clause_map[*it]-1));
    clause_indexes.push_back(*it);

//    if (debug)
//      std::cout << "DEBUG: Added [" << *it << "] (" << clause_map[*it] <<
//        ") = " << qcnf.at(clause_map[*it]-1).to_string() << std::endl;
  }

  antecedents.clear();

  unsigned last_dirty = 0;
  unsigned cur = 0;
  if (clauses.size()!=0)
  do
  {
    for (DClause::const_iterator rit = assumptions.begin();
         rit!=assumptions.end();
         rit++)
    {
      Literal l = *rit; // assume -l is true
      clauses[cur].remove(l);
      l.invert();
//      if (clauses[cur].contains(l))
//      {
//        //std::cout << "Satisfied clause" << std::endl;
//        goto next;
//      }
    }

    unsigned virtual_clause_size=0;
    Literal virtual_unit(0,0);

    for (DClause::const_iterator cit = clauses[cur].begin();
         cit!=clauses[cur].end();
         cit++)
    {
      if (qcnf.quantifiers[cit->variable()].type==Quantifier::EXISTENTIAL)
      {
        virtual_clause_size++;
        virtual_unit = *cit;
      }
    }

    if (virtual_clause_size==0)
    {
      antecedents.push_back(clause_indexes[cur]);
    }
    else
    {
      virtual_unit.invert();
      if (virtual_clause_size==1 &&
          assumptions.insert(virtual_unit))    // new unit
      {
        // std::cout << "Found assumption: " << u.to_dimacs() << std::endl;
        antecedents.push_back(clause_indexes[cur]);
        pivots.push_back(virtual_unit.variable());
        last_dirty = cur;
      }
    }

//    next:
    cur++;
    if (cur==clauses.size()) cur = 0; //wrap
  } while (cur!=last_dirty);

  if (antecedents.size()==checksize-1)
  {
    if (cur==0) //one back
      cur = clauses.size()-1;
    else
      cur--;
    antecedents.push_back(clause_indexes[cur]);
  }
  else if (antecedents.size()!=checksize)
  {
    for (std::list<unsigned>::const_iterator it = antecedents.begin();
         it!=antecedents.end();
         it++)
    {
      if (*it>=clause_map.size() || clause_map[*it]==0)
        throw Exception("Inconsistent clause map.");
      std::cout << *it << ": " << qcnf.at(clause_map[*it]-1).to_string() <<
        std::endl;
    }
    throw Exception("BUG: Lost antecedents while reordering.");
  }

  antecedents.reverse();
  pivots.reverse();

  if (debug)
  {
//    std::cout << "DEBUG: new order: ";
//    for (std::list<unsigned>::const_iterator it = antecedents.begin();
//         it!=antecedents.end();
//         it++)
//    {
//      std::cout << *it << ", ";
//    }
//    std::cout << std::endl;
//    std::cout << "DEBUG: pivots: ";
//    for (std::list<unsigned>::const_iterator it = pivots.begin();
//         it!=pivots.end();
//         it++)
//    {
//      std::cout << *it << ", ";
//    }
//    std::cout << std::endl;
  }
}

/*******************************************************************\

Function: QBV::show_model

  Inputs: model

 Outputs: none

 Purpose: prints the model functions

\*******************************************************************/

void QBV::show_model(model_type &model) const
{
  std::cout << "Model functions: " << std::endl;

  for(model_type::const_iterator it=model.begin();
      it!=model.end();
      it++)
  {
    std::cout << it->first << " = ";
    show_model_function(model, it->second);
    std::cout << std::endl;
  }
}

/*******************************************************************\

Function: QBV::show_model_function

  Inputs: model, function id

 Outputs: none

 Purpose: prints a model function

\*******************************************************************/

void QBV::show_model_function(model_type &model, signed f) const
{
  Literal l( (f<0)?-f:f, (f<0) );
  unsigned var = l.variable();

  if(qcnf.is_universal(var))
    std::cout << l.to_dimacs();
  else
  {
    if(l.sign()) std::cout << "!";

    model_type::const_iterator model_it=model.find(var);
    model_cache::const_iterator ite_it=model_cache_ite.find(var);
    model_cache::const_iterator and_it=model_cache_and.find(var);

    if(model_it!=model.end())
      show_model_function(model, model_it->second);
    else if(ite_it!=model_cache_ite.end())
    {
      const std::vector<Literal> &m=ite_it->second;
      std::vector<Literal>::const_iterator mit=m.begin();

      assert(m.size()==3);

      Literal cond=(*mit++);
      Literal tb=(*mit++);
      Literal fb=(*mit++);

      std::cout << "(";

      if(cond==tb)
      {
        show_model_function(model, cond.to_dimacs());
        std::cout << " || ";
        show_model_function(model, fb.to_dimacs());
      }
      else if(cond==fb)
      {
        show_model_function(model, (~cond).to_dimacs());
        std::cout << " || ";
        show_model_function(model, tb.to_dimacs());
      }
      else
      {
        std::cout << "(";
        show_model_function(model, cond.to_dimacs());
        std::cout << ")?";
        show_model_function(model, tb.to_dimacs());
        std::cout << ":";
        show_model_function(model, fb.to_dimacs());
      }

      std::cout << ")";
    }
    else if(and_it!=model_cache_and.end())
    {
      const std::vector<Literal>& m=and_it->second;

      if(m.size()==0)
        std::cout << "true";
      else
      {
        if(m.size()>1)
          std::cout << "(";

        for(std::vector<Literal>::const_iterator it=m.begin();
            it!=m.end();
            it++)
        {
          if(it!=m.begin()) std::cout << " && ";
          show_model_function(model, it->to_dimacs());
        }

        if(m.size()>1)
          std::cout << ")";
      }
    }
    else
    {
      assert(and_it==model_cache_and.end() &&
             ite_it==model_cache_ite.end());
      std::cout << "*";
    }
  }
}

/*******************************************************************\

Function: QBV::apply_delayed_extensions

  Inputs:

 Outputs:

 Purpose: Applies as many saved extensions as possible

\*******************************************************************/

void QBV::apply_delayed_extensions(void)
{
  if (debug)
    std::cout << "Applying delayed extensions ..." << std::endl;
  
  and_ext_list.reverse();
  ite_ext_list.reverse();

  bool progress=true;
  while(progress)
  {
    progress=false;

    for(and_extension_list::iterator it=and_ext_list.begin();
        it!=and_ext_list.end();
        /* nothing */)
    {
      and_extension &ext = *it;

      if(apply_extension_and(ext))
      {
        progress=true;
        it = and_ext_list.erase(it);
      }
      else
        it++;
    }

    for(ite_extension_list::iterator it=ite_ext_list.begin();
        it!=ite_ext_list.end();
        /* nothing */)
    {
      ite_extension &ext = *it;

      if(apply_extension_ite(ext))
      {
        progress=true;
        it = ite_ext_list.erase(it);
      }
      else
        it++;
    }
  }
}

/*******************************************************************\

Function: QBV::apply_delayed_resolutions

  Inputs:

 Outputs:

 Purpose: Applies as many saved resolutions as possible

\*******************************************************************/

void QBV::apply_delayed_resolutions(void)
{
  if(debug)
    std::cout << "Applying delayed resolutions ..." << std::endl;;

  bool progress=true;
  while (progress)
  {
    progress=false;

    for (std::list<Resolution>::iterator it = resolvents_list.begin();
         it!=resolvents_list.end();
         /* Nothing */)
    {
      Resolution &r = *it;
      bool all_antecedents_here = true;
      for (std::list<unsigned>::const_iterator ait = r.antecedents.begin();
           ait!=r.antecedents.end();
           ait++)
      {
        if (*ait>=clause_map.size() || clause_map[*ait]==0)
          { all_antecedents_here = false; break; }
      }

      if (all_antecedents_here)
      {
        Resolution copy = r;
        it = resolvents_list.erase(it); // invalidates r;
        progress = true;
        apply_resolution(copy);
      }
      else
        it++;
    }
  }
}

/*******************************************************************\

Function: QBV::check_delayed_resolutions

  Inputs: none

 Outputs: none

 Purpose: Checks for missing resolvents

\*******************************************************************/

void QBV::check_delayed_resolutions(void) const
{
  if (!resolvents_list.empty())
  {
    std::stringstream s;
    s << "The following delayed resolutions are missing antecedents " <<
      "at conclusion: " << std::endl;
    for (std::list<Resolution>::const_iterator it = resolvents_list.begin();
         it!=resolvents_list.end();
         it++)
    {
      s << it->newinx << ": ";
      for (std::list<unsigned>::const_iterator ait = it->antecedents.begin();
           ait!=it->antecedents.end();
           ait++)
      {
        s << *ait;
        if (clause_map[*ait]==0) s << "*";
        s << ", ";
      }
      s << std::endl;
    }
    throw Exception(s.str());
  }
}

/*******************************************************************\

Function: QBV::check_model_quantification

  Inputs: none

 Outputs: none

 Purpose: Checks if a model function depends on variables that are
          not in its scope.

\*******************************************************************/

void QBV::check_model_quantification(model_type &model) const
{
  for(model_type::const_iterator it=model.begin();
      it!=model.end();
      it++)
  {
    unsigned var=it->first;
    
    if (qcnf.quantifiers[var].type!=Quantifier::EXISTENTIAL)
      throw Exception("Universal or unquantified variable in model.");
    
//    if (qcnf.is_before(it->first, m.variable()))
//    {
//      std::stringstream s;
//      s << "Model function for " << it->first << " (on lvl " <<
//        qcnf.quantification_levels[it->first] << ") depends on variables " <<
//        "with higher quantification (" << m.variable() << " on lvl " <<
//        qcnf.quantification_levels[m.variable()] << ").";
//      throw Exception(s.str());
//    }    
  }
}

/*******************************************************************\

Function: QBV::check_delayed_extensions

  Inputs: none

 Outputs: none

 Purpose: Checks for missing extensions

\*******************************************************************/

void QBV::check_delayed_extensions(void) const
{
  if(and_ext_list.size()>0 || ite_ext_list.size()>0)
  {
    std::stringstream s;
    s << "The following extensions were not applied because of missing " <<
      "dependencies at conclusion:" << std::endl;

    for(and_extension_list::const_iterator it=and_ext_list.begin();
        it!=and_ext_list.end();
        it++)
    {
      s << "E " << it->new_var.to_dimacs() << " (line #" << it->line_no << ")"
        << std::endl;
    }

    for(ite_extension_list::const_iterator it=ite_ext_list.begin();
        it!=ite_ext_list.end();
        it++)
    {
      s << "E " << it->new_var.to_dimacs() << " (line #" << it->line_no << ")"
        << std::endl;
    }

    throw Exception(s.str());
  }
}

/*******************************************************************\

Function: QBV::check_function_dependencies

  Inputs: model

 Outputs: none

 Purpose: Checks for cyclic dependencies in model functions

\*******************************************************************/

void QBV::check_function_dependencies(model_type &model)
{  
  if(debug)
    std::cout << "Checking function dependencies ..." << std::endl;

  // This is an implementation of a topological sort.
  // The vector `no_incoming' must be empty after sorting,
  // otherwise there is a cyclic dependency in the model.
  // While we sort, we also record the set of universals
  // that every existential depends on (in `universal_dependencies').
    
  dependency_map_type incoming_edges;
  dependency_map_type outgoing_edges;
  dependency_stack_type no_incoming;
  
  // Map: variable -> (variable, level) representing max. lvl dependency.
  typedef std::map<unsigned, std::pair<unsigned, unsigned> > max_universal_type;
  max_universal_type universal_dependencies;
  
  // First, we build the dependency graph (incoming/outgoing_edges). 
  construct_dependency_graph(model, incoming_edges, outgoing_edges, no_incoming);
  
  function_order.reserve(qcnf.max_variable_no+1);

  // Then we sort.
  while(!no_incoming.empty())
  {
    unsigned v=no_incoming.back();
    no_incoming.pop_back();
    function_order.push_back(v);

    assert(incoming_edges.find(v)==incoming_edges.end());

    std::map<unsigned, dependency_set_type>::iterator ofit=outgoing_edges.find(v);

    if(ofit!=outgoing_edges.end())
    {
      dependency_set_type &outset=ofit->second;

      for(dependency_set_type::iterator out_it=outset.begin();
          out_it!=outset.end();
          out_it++)
      {
        unsigned ov = *out_it;
        
        // Update the maximum universal dependency level.
        // (ov depends on v)
        if(qcnf.is_universal(v))
        {
          if(qcnf.quantification_levels[v] > universal_dependencies[ov].second)
            universal_dependencies[ov]=
                std::make_pair(v,qcnf.quantification_levels[v]);
        }
        else
        {
          max_universal_type::const_iterator 
              v_it=universal_dependencies.find(v);
          
          if(v_it!=universal_dependencies.end() &&
             v_it->second.second > universal_dependencies[ov].second)
            universal_dependencies[ov]=v_it->second;          
        }
        
        // Go on with the sorting

        std::map<unsigned, dependency_set_type >::iterator fit =
          incoming_edges.find(ov);

        if (fit==incoming_edges.end())
          no_incoming.push_back(ov);
        else
        {
          dependency_set_type &inset = fit->second;

          unsigned i=0;
          for(; i<inset.size(); i++)
            if(inset[i]==v) { inset[i]=0; break; }
          assert(i<inset.size());

          while(inset.size()>0 && inset[inset.size()-1]==0)
            inset.pop_back();

          if(inset.empty())
          {
            no_incoming.push_back(ov);
            incoming_edges.erase(fit);
          }
        }
      }
      
      outgoing_edges.erase(ofit);
    }
  }

  // The check for cyclic dependencies.
  if(!incoming_edges.empty())
  {
    if(debug)
    {
      std::cout << "Remaining edges:" << std::endl;
      for(std::map<unsigned, dependency_set_type >::const_iterator it=incoming_edges.begin();
          it!=incoming_edges.end();
          it++)
      {
        std::cout << it->first << ": ";
        bool first=true;
        for(dependency_set_type::const_iterator sit=it->second.begin();
            sit!=it->second.end();
            sit++)
        {          
          if(*sit!=0) 
          {
            if(!first) std::cout << ", ";
            std::cout << *sit;
            first=false;
          }
        }
        std::cout << std::endl;;
      }
    }
    throw Exception("The model provided contains cyclic function definitions.");
  }
  
  // At last, we check what universals each existential depends on.
//  std::cout << "Universal dependencies: " << std::endl;
  for(max_universal_type::const_iterator it=universal_dependencies.begin();
      it!=universal_dependencies.end();
      it++)
  {
    unsigned v=it->first;
    const std::pair<unsigned, unsigned> &max_dep=it->second;
    
//    std::cout << v << ": " << max_dep.first << 
//      " (" << max_dep.second << ")" << std::endl;
        
    if(max_dep.first!=0 && !qcnf.is_before(max_dep.first, v))
    {
      std::stringstream s;
      s << "Model function for " << v << " (on level " <<
        qcnf.quantification_levels[v] << ") depends on universal " <<
        "variable with higher quantification (" << max_dep.first << " on level " <<
        max_dep.second << ").";
      throw Exception(s.str());
    }        
  }
}

/*******************************************************************\

Function: QBV::construct_dependency_graph

  Inputs: model

 Outputs: none

 Purpose: Constructs the variable dependency graph 

\*******************************************************************/

void QBV::construct_dependency_graph(
  const model_type &model,
  dependency_map_type &incoming_edges,
  dependency_map_type &outgoing_edges,
  dependency_stack_type &no_incoming) const
{
  // Add edges for ITE extension dependencies
  for(model_cache::const_iterator it=model_cache_ite.begin();
      it!=model_cache_ite.end();
      it++)
  {
    const std::vector<Literal> &m=it->second;
    assert(m.size()==3);

    incoming_edges[it->first].reserve(3);
    incoming_edges[it->first].push_back(m[0].variable());
    incoming_edges[it->first].push_back(m[1].variable());
    incoming_edges[it->first].push_back(m[2].variable());

    outgoing_edges[m[0].variable()].push_back(it->first);
    outgoing_edges[m[1].variable()].push_back(it->first);
    outgoing_edges[m[2].variable()].push_back(it->first);
  }

  // Add edges for AND extension dependencies
  for(model_cache::const_iterator it=model_cache_and.begin();
      it!=model_cache_and.end();
      it++)
  {
    const std::vector<Literal>& m=it->second;

    if(m.size()>0)
      incoming_edges[it->first].reserve(m.size());

    for(std::vector<Literal>::const_iterator ait=m.begin();
        ait!=m.end();
        ait++)
    {
      incoming_edges[it->first].push_back(ait->variable());
      outgoing_edges[ait->variable()].push_back(it->first);
    }
  }

  // Add edges for the model dependencies
  for(std::vector<QCNF::qscopet>::const_iterator it=qcnf.quantification_sequence.begin();
      it!=qcnf.quantification_sequence.end();
      it++)
    for(std::list<Quantifier>::const_iterator qit=it->begin();
        qit!=it->end();
        qit++)
    {
      unsigned v=qit->variable;

      if(qcnf.is_universal(v))
        no_incoming.push_back(v);
      else
      {
        model_type::const_iterator mit=model.find(v); 
        if(mit!=model.end())
        {
          incoming_edges[v].push_back(Literal(mit->second).variable());
          outgoing_edges[Literal(mit->second).variable()].push_back(v);
        }
        else
        { // not a model ...
          if(incoming_edges.find(v)==incoming_edges.end()) // and not an extension
            no_incoming.push_back(v);
        }
      }
    }
}
