/*******************************************************************\

Module: QBV, The Quantified Boolean Validator

Author: CM Wintersteiger

\*******************************************************************/

#ifndef QBV_H_
#define QBV_H_

#include <map>
#include <set>
#include <vector>
#include <istream>
#include <string>

#include <time.h>

#ifdef __MINGW32__
#include <sys/time.h>
#endif

#include "cnf.h"
#include "extensions.h"

class QBV {
  public:
    unsigned cur_linenumber, cur_colnumber;
    QBV( void );
    
    // Validation functions
    bool validate(const std::string &, const std::string &);
    bool validate(const char *, const char *);
    bool validate_trace(const std::string &);
    bool validate_trace(const char *);
    
    // Some options
    void set_test_trace( void ) { test_trace=true; }
    void set_ignore_no_reason( void ) { ignore_no_reason=true; }
    void set_debug( bool v ) { debug = v; }
    
  private:
    QCNF qcnf;
    unsigned last_original_clause;
    unsigned last_original_variable;
    
    bool  test_trace,
          debug,
          ignore_no_reason;
    
    std::string cur_token;
    
    char read_character( std::ifstream & );
    std::string& read_token( std::ifstream & );
    unsigned read_unsigned( std::ifstream & );
    signed read_signed( std::ifstream & );
    void read_whitespace( std::ifstream & );
    Literal read_literal( std::ifstream & );
        
    bool read_conclusion_line( std::ifstream & );
    
    // Extensions
    typedef std::list<and_extension> and_extension_list;
    typedef std::list<ite_extension> ite_extension_list;
    and_extension_list and_ext_list;
    ite_extension_list ite_ext_list;
    
    void read_extension_line( std::ifstream & );
    
    and_extension read_extension_and(unsigned, std::ifstream & );
    ite_extension read_extension_ite(unsigned, std::ifstream & );
    
    bool apply_extension_ite( ite_extension & );
    bool apply_extension_and( and_extension & );    
    void apply_delayed_extensions(void);    
    void check_delayed_extensions(void) const;
    
    // Resolution functions
    class Resolution {
      public:
        bool star_resolvent;
        unsigned newinx; 
        DClause resolvent;
        std::list<unsigned> antecedents; 
    };
    
    std::list<Resolution> resolvents_list;
    std::vector<unsigned> clause_map;
    
    void read_resolution_line( std::ifstream & );
    void apply_resolution( Resolution & );
    void apply_delayed_resolutions(void);
    void check_delayed_resolutions(void) const;
    
    void order_antecedents( const DClause &, 
                           std::list<unsigned> &,
                           std::list<unsigned> &) const;        
    
    // Model functions
    std::vector<unsigned> function_order;
    class model_cache : public std::map<unsigned, std::vector<Literal> > {};
    model_cache model_cache_ite;
    model_cache model_cache_and;
    
    typedef std::map<unsigned, signed> model_type;
    
    void check_model_quantification(model_type &model) const;
    
    void show_model(model_type&) const;
    void show_model_function(model_type&, signed) const;
    
    bool test_clausewise(model_type&);
    
    // Trace checking   
    void read_trace_line(unsigned, std::ifstream &);
    
    // Model dependencies
    typedef std::vector<unsigned> dependency_set_type;
    typedef std::map<unsigned, dependency_set_type > dependency_map_type;
    typedef std::vector<unsigned> dependency_stack_type;    
        
    void check_function_dependencies(model_type &model);
        
    void construct_dependency_graph(
      const model_type&, dependency_map_type&, 
      dependency_map_type&, dependency_stack_type&) const;
      
    struct timeval starttime;    
};

#endif
